/*
 * libsapilektor.c - SAPI interface
 * Copyright (C) Bohdan R. Rau 2012 <ethanak@polip.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <iconv.h>
#include <stdarg.h>

#include <fcntl.h>
#include <errno.h>
#include "libsapilektor.h"

const char * const SAPI_version(void)
{
    return SAPI_VERSION_B;
}

static char *trim(char *str);

#include "sapi_config.c"

#define SAPI_MODE_NATIVE 1
#define SAPI_MODE_RBE 2
#define SAPI_MODE_S4L 3


#define SAPI_FRAME_LENGTH 256

struct SAPI_PrivData {
    struct SAPI_data public;
    struct sapiConf *conf;
    struct sapiSection *sect;
    int mode;

    int timeout;

    struct sockaddr_in rockbox_sinadr; // also for s4l
    int rockbox_port; //also for s4l
    int rockbox_ethanak;
    char *rockbox_output_dir; // dla ethanak=0
    int rockbox_voice_id;
    int rockbox_autospawn;
    char *rockbox_spawn_dir;
    char *rockbox_spawn_cmd;
    char *rockbox_wine;
    char *s4l_voice;
    int ssml_safe;
    FILE *debug_file;
};


static void _dbg(void *s,char *fmt,...)
{
    FILE *f=((struct SAPI_PrivData *)s)->debug_file;
    if (!f) return;
    va_list vl;
    va_start(vl,fmt);
    vfprintf(f,fmt,vl);
    va_end(vl);
    fflush(f);
}

#include "sapi_network.c"
#include "sapi_rockbox.c"
#include "sapi_s4l.c"

static struct SAPI_data *sapi_open_network(char *name,char *host,int port,char **error)
{
    struct sapiSection *ise;
    char *pm;
    struct sapiConf *conf=sapi_ReadConfig("sapilektor.cfg");
    if (!conf) {
        if (error) *error=strdup(strerror(errno));
        return NULL;
    }
    if (conf->ErrorStatus) {
        if (error) {
            *error=malloc(strlen(conf->Error)+128);
            sprintf(*error,"%s in line %d",conf->Error,conf->ErrorLine);
        }
        sapi_FreeConf(conf);
        return NULL;
    }
    for (ise=conf->sections;ise;ise=ise->next) {
        if (!strcasecmp(ise->name,name)) break;
    }
    if (!ise) {
        sapi_FreeConf(conf);
        if (error) {
            *error=malloc(strlen(name)+128);
            sprintf(*error,"No section [%s] in config",name);
        }
        return NULL;
    }
    pm=sapi_ConfParam(ise,"mode");
    if (!pm) {
        if (error) {
            *error=malloc(strlen(name)+128);
            sprintf(*error,"No 'mode' parameter in section [%s]",name);
        }
        sapi_FreeConf(conf);
        return NULL;
    }
    if (!strcasecmp(pm,"native")) {
	if (error) {
            *error=malloc(strlen(name)+128);
            sprintf(*error,"IVONA Native mode not implemented in section [%s]",name);
        }
        sapi_FreeConf(conf);
        return NULL;
        
    }
    if (!strcasecmp(pm,"opensapi.rockbox")) {
        return sapiOpenRockbox(conf,ise,host,port,error);
    }
    if (!strcasecmp(pm,"sapi4linux")) {
        return sapiOpenS4l(conf,ise,host,port,error);
    }
    
    if (error) {
        *error=malloc(strlen(name)+strlen(pm)+128);
        sprintf(*error,"Unknown mode '%s' in section [%s]",pm,name);
    }
    sapi_FreeConf(conf);
    return NULL;
}
struct SAPI_data *SAPI_open(char *name,char **error)
{
    return sapi_open_network(name,NULL,0,error);
}

struct SAPI_data *SAPI_openNetwork(char *name,char *host,int port,char **error)
{
    return sapi_open_network(name,host,port,error);
}

void SAPI_enableSSML(struct SAPI_data *handle,int enable)
{
	((struct SAPI_PrivData *)handle)->ssml_safe=enable;
}


static int uni_isspace(int znak)
{
	int i;
	static int spacje[]={0x1680,0x180e,0x2028,0x2029,0x205f,0x3000,0};
	if (znak<=0x20) return 1;
	if (znak>=0x7f && znak <=0xa0) return 1;
	if (znak>=0x2000 && znak <=0x200b) return 1;
	for (i=0;spacje[i];i++) if (znak==spacje[i]) return 1;
	return 0;
}

static int get_unichar(char **str)
{
	int znak,n,m;
	if (!*str) return 0;
	znak=(*(*str)++) & 255;
	if (!(znak & 0x80)) return znak;
	if ((znak & 0xe0)==0xc0) n=1;
	else if ((znak & 0xf0)==0xe0) n=2;
	else return 0;
	znak &= 0x1f;
	while (n--) {
		m=*(*str)++ & 255;
		if ((m & 0xc0)!=0x80) {
			return 0;
		}
		znak=(znak<<6) | (m & 0x3f);
	}
	return znak;
}


static double sapi_compute_frame(short *w)
{
    int j;double sum,d;
    sum=0.0;
    for (j=0;j<SAPI_FRAME_LENGTH;j++) {
        d=w[j]/32767.0;
        if (d<0) d=-d;
        sum+=d;
    }
    return sum/SAPI_FRAME_LENGTH;
}

static void normalize_wave(short *wave,int slength)
{
    int mxn,i;
    for (i=mxn=0;i<slength;i++) {
        if (wave[i]>mxn) mxn=wave[i];
        else if (-wave[i]>mxn) mxn=-wave[i];
    }
    if (mxn < 32700 && mxn > 100) {
        double mpx=32767.0/mxn;
        for (i=0;i<slength;i++) wave[i] *= mpx;
    }
}

static short *sapi_getWave(struct SAPI_PrivData *handle,char *text,int *length,char **error)
{
    int slength,mxn,i;
    short *wave;
    _dbg(handle,"getWave entered with text '%s'\n",text);
    if (!handle->ssml_safe && handle->mode != SAPI_MODE_S4L) {
		while (*text && (isspace(*text) || *text=='<')) text++;
	}
    switch (handle->mode) {
        case SAPI_MODE_RBE:
        wave=sapi_RockboxGetWave(handle,text,&slength,error);
        break;

        case SAPI_MODE_S4L:
        wave=sapi_S4lGetWave(handle,text,&slength,error);
        break;

        default:
        if (error) *error=strdup("Bad MODE");
        return NULL;
    }
    _dbg(handle,"wave got %p size=%d\n",wave,slength);
    if (!wave) return NULL;
    normalize_wave(wave,slength);
    _dbg(handle,"wave_normalized\n");
    *length=slength;
    return wave;
}


static int sapi_trimWave(struct SAPI_PrivData *handle,short *wave,int length,int *offset,int *rest)
{
    int i,ofs,rst;double sum;
    /* w handle można trzymać np. wartości odcięć */
    ofs=-1;rst=-1;
    for (i=0;i<length-SAPI_FRAME_LENGTH;i+=SAPI_FRAME_LENGTH) {
        sum=sapi_compute_frame(wave+i)  * handle ->public.silmpx ;
        if (sum >= 0.03) {
            ofs=i;
            break;
        }
    }
    if (ofs < 0) return -1;
    for (i=length-SAPI_FRAME_LENGTH;i>ofs;i-=SAPI_FRAME_LENGTH) {
        sum=sapi_compute_frame(wave+i)  * handle ->public.silmpx ;
        if (sum >= 0.05) {
            rst=i+SAPI_FRAME_LENGTH;
            break;
        }
    }
    if (rst < 0) return -1;
    *offset=ofs;
    *rest=length-rst;
    return 0;
}

static int sapi_trimLastPart(struct SAPI_PrivData *handle,
    short *wave,
    int length,
    int *out_length,
    int *rest)
{
    int i,phase,rlen,slen;double sum;
    for (i=length-SAPI_FRAME_LENGTH,phase=0;i>0;i-=SAPI_FRAME_LENGTH) {
        sum=sapi_compute_frame(wave+i) * handle ->public.silmpx;
        if (!phase) {
            if (sum > 0.1) phase=1;
            continue;
        }
        if (phase == 1) {
            if (sum < 0.01) {
                rlen=i;
                phase=2;
            }
            continue;
        }
        if (sum >=0.05) {
            phase=3;
            slen=i+SAPI_FRAME_LENGTH;
            break;
        }
    }
    if (phase != 3) return -1;
    *out_length=slen;
    *rest=rlen-slen;
    return 0;
}


/* publiczne funkcje */
static short *sapi_repauser(struct SAPI_PrivData *handle,short *wave,int *len,double factor,int *offset,int *rest);


void SAPI_debugFile(struct SAPI_data *handle,FILE *file)
{
    ((struct SAPI_PrivData *)handle)->debug_file=file;
}

short *SAPI_getSentence(struct SAPI_data *handle,
    char *sentence,
    int *sample_offset,
    int *sample_count,
    int *sample_rest,
    double pause_factor,
    char **error)
{
    short *wave,*wave2;
    int soffset,srest,slength;
    wave=sapi_getWave((struct SAPI_PrivData *)handle,sentence,&slength,error);
    if (!wave) return NULL;

    soffset=0;
    srest=0;
    wave2=sapi_repauser((struct SAPI_PrivData *)handle,wave,&slength,pause_factor,&soffset,&srest);
    if (wave2) {
    	free(wave);
	wave=wave2;
    }
    *sample_offset=soffset;
    *sample_count=slength-(soffset+srest);
    *sample_rest=srest;
    return wave;
}

short *SAPI_getWave(struct SAPI_data *handle,char *text,int *length,char **error)
{
    return sapi_getWave((struct SAPI_PrivData *)handle,text,length,error);
}

int SAPI_trimWave(struct SAPI_data *handle,short *wave,int length,int *offset,int *rest)
{
    return sapi_trimWave((struct SAPI_PrivData *)handle,wave,length,offset,rest);
}


int SAPI_trimLastPart(struct SAPI_data *handle,
    short *wave,
    int length,
    int *out_length,
    int *rest)
{
    return sapi_trimLastPart((struct SAPI_PrivData *)handle,wave,length,out_length,rest);
}

void SAPI_close(struct SAPI_data *handle)
{
    sapi_FreeConf(((struct SAPI_PrivData *)handle)->conf);
    switch (((struct SAPI_PrivData *)handle)->mode) {
        case SAPI_MODE_RBE:
        sapi_RockboxClose((struct SAPI_PrivData *)handle);
        break;

        case SAPI_MODE_S4L:
        sapi_S4lClose((struct SAPI_PrivData *)handle);
        break;
    }
    free(handle);
}

void *SAPI_initWave(struct SAPI_data *handle,char *text,char **error)
{
    switch (((struct SAPI_PrivData *)handle)->mode) {
        case SAPI_MODE_RBE:
        return (void *)sapi_RockboxInitWave((struct SAPI_PrivData *)handle,text,error);
        case SAPI_MODE_S4L:
        return (void *)sapi_S4lInitWave((struct SAPI_PrivData *)handle,text,error);
        default:
        if (error) *error=strdup("Bad MODE");
        return NULL;
    }
    return NULL;
}

short *SAPI_fetchWave(struct SAPI_data *handle,void *f,int *sample_length,char **error)
{
    switch (((struct SAPI_PrivData *)handle)->mode) {
        case SAPI_MODE_RBE:
        return (void *)sapi_RockboxFetchWave((struct SAPI_PrivData *)handle,(FILE *)f,sample_length,error);
        case SAPI_MODE_S4L:
        return (void *)sapi_S4lFetchWave((struct SAPI_PrivData *)handle,(FILE *)f,sample_length,error);
        default:
        if (error) *error=strdup("Bad MODE");
        return NULL;
    }
    return NULL;
}

void SAPI_freeWaveHandler(struct SAPI_data *handle,void *hd)
{
    switch (((struct SAPI_PrivData *)handle)->mode) {

	case SAPI_MODE_RBE:
	sapi_RockboxFreeWHD((FILE *)hd);
	break;

	default:
	break;
    }
}
#include "sapi_pauser.c"
