/*
 * txt2wav - utility libsapilektor
 * Copyright (C) Bohdan R. Rau 2012 <ethanak@polip.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libsapilektor.h>
#include <sndfile.h>
#include <getopt.h>
#include <sys/stat.h>
#include <glib.h>
#include <unistd.h>
#include <fcntl.h>
#include "sonic.h"
#include "audio.h"

double pause_factor = 1.0;
double speed = 1.0;
char *voice=NULL;
char **voices;
char *dialog=NULL;
int audiobook=0;
int titleline=0;
char *endline=NULL;
double breath=1.0;
int is_raw=0;
int do_play;
int autodivide;
int sentenced;

#define PAR_EMPTY 1
#define PAR_NORMAL 2
#define PAR_DIALOG 3

#define PAUSE_NORMAL 1
#define PAUSE_LONG 2
#define PAUSE_PRE 3
#define PAUSE_POST 4
#define PAUSE_VLONG 5
#define PAUSE_SENTENCE 6

int _pauses[]={0,800,1600,1100,1250,2500,100};

void helpme(char *program, int estatus)
{
    int i;
    fprintf(stderr,"SAPI lektor v %s\n",SAPI_version());
    fprintf(stderr,"Usage: %s [options] [infile.txt [outfile.wav]]\n",program);
    fprintf(stderr,"Options are:\n\
  -p <float> - pause factor (0.2 .. 1.0)\n\
  -b <float> - breath factor (0.25 .. 2.0)\n\
  -s <float> - speed (0.5 .. 2.0)\n\
  -v <name>  - voice name (");
    for (i=0;voices[i];i++) fprintf(stderr,"%s%s",i?", ":"",voices[i]);

   fprintf(stderr,")\n\
  -a         - audiobook mode\n\
  -t         - first line is chapter title (implies -a)\n\
  -e <text>  - announce end of chapter/book at end of text (implies -a)\n\
  -d <char>  - dialog list marker (implies -a)\n\
  -x         - trailing backslash means end of sentence, not paragraph\n\
               (excludes -y)\n\
  -y         - automatically divide into sentences (excludes -x)\n\
  -r         - create raw audio instead of wav\n\
  -o         - don't create audio file, play the sound instead\n\
  -l         - enable inline SSML for opensapi.rockbox (use with caution!)\n\
  -f         - input is single SSML document\n\
  -g <path>  - debug output file\n\
  -w         - flush output wave after each paragraph\n\
  -h         - this help\n");
   exit(estatus);
}

FILE *fdebug=NULL;
void _dbg(char *format,...)
{
    va_list ar;
    if (!fdebug) return;
    va_start(ar,format);
    vfprintf(fdebug,format,ar);
    va_end(ar);
    fflush(fdebug);
}

struct paragraph {
    int par_pause;
    char *par_body;
} *pars;
int para_count,para_size=0;
char *body;

struct SAPI_data *vox;

static SNDFILE *output_sf;
static sonicStream sonic;



void open_wave(int freq,char *fname)
{
    SF_INFO info;
    
    char *c;int ogg=0;
    

    if (speed != 1.0) {
	_dbg("Creating sonic stream for rate = %.2f\n",speed);
        sonic=sonicCreateStream(freq,1);
        sonicSetSpeed(sonic,speed);
    }

    if (do_play) {
	_dbg("Initializing libao with freq=%d\n",freq);
        init_ao(freq);
        return;
    }
    
    
    memset(&info,0,sizeof(info));
    if (!fname) fname="-";
    else {
		c=strrchr(fname,'.');
		if (c && !strcasecmp(c,".ogg")) ogg=1;
	}
    _dbg("Opening output %d sps stream %s\n",freq,fname);
    info.samplerate=freq;
    info.channels=1;
    if (ogg) info.format=SF_FORMAT_OGG | SF_FORMAT_VORBIS;
    else info.format=(is_raw ? SF_FORMAT_RAW : SF_FORMAT_WAV) | SF_FORMAT_PCM_16 | SF_ENDIAN_LITTLE;
    if (!strcmp(fname,"-")) {
        output_sf=sf_open_fd(1,SFM_WRITE,&info,0);
    }
    else {
        output_sf=sf_open(fname,SFM_WRITE,&info);
    }
    if (!output_sf) {
        perror(fname);
        exit(1);
    }
}

void wreal_wave(short *wave,int nsamples)
{
    if (do_play) {
        play_audio(wave,nsamples);
        return;
    }
    if (sf_write_short(output_sf,wave,nsamples)!=nsamples) {
        perror("Wave write");
        exit(1);
    }
    _dbg("Wave written %d samples\n",nsamples);
}

void write_wave(short *wave,int nsamples)
{
    if (!nsamples) {
	_dbg("Zero wave length, ignored\n");
	return;
    }
    if (sonic) {
        short buf[32768];
        int n,inlen;
        while (nsamples > 0) {
            inlen=16384;
            if (inlen > nsamples) inlen=nsamples;
            sonicWriteShortToStream(sonic,wave,inlen);
            wave+=inlen;
            nsamples-=inlen;
            for (;;) {
                n=sonicReadShortFromStream(sonic,buf,32768);
                if (n<=0) break;
                wreal_wave(buf,n);
            }
        }
        return;
    }
    wreal_wave(wave,nsamples);
}

void finalize_wave(void)
{
    if (do_play) {
        audio_close_device();
        return;
    }
    sf_close(output_sf);
}

void sync_wave(void)
{
    if (!do_play) sf_write_sync(output_sf);
}

void push_wave(short *wave,int offset,int count)
{
    wave += offset;
    write_wave(wave,count);
}

void push_silence(int pause)
{
    static short buf[1024];
    static int w=0;
    int n;
    if (!w) {
        memset(buf,0,2048);
        w=1;
    }
    while(pause > 0) {
        n=1024;
        if (n > pause) n=pause;
        push_wave(buf,0,n);
        pause -= n;
    }
}

void push_pause(short *w1,int l1,short *w2,int l2,int pause)
{
    int m1,m2;
    if (l1 + l2 <= pause) {
	_dbg("pause %d - %d - %d\n",l1,pause-(l1+l2),l2);
        push_wave(w1,0,l1);
        push_silence(pause-(l1+l2));
        push_wave(w2,0,l2);
        return;
    }
    m1=(pause * l1)/(l1+l2);
    m2=pause-m1;
    _dbg("pause %d/%d : %d/%d\n",m1/l1,m2/l2);
    push_wave(w1,0,m1);
    push_wave(w2,l2-m2,m2);
}

char *trim(char *c)
{
    gunichar z;
    int was=0;
    char *d,*e;
    
    while (*c) {
        z=g_utf8_get_char(c);
        if (!g_unichar_isspace(z)) break;
        c=g_utf8_next_char(c);
    }
    if (!*c) return c;
    for (d=e=c;*d;) {
        z=g_utf8_get_char(d);
        d=g_utf8_next_char(d);
        if (!g_unichar_isspace(z)) e=d;
        if (g_unichar_isalnum(z)) was=1;
    }
    if (!was) e=c;
    *e=0;
    return c;
}

int is_dialog(char *dialog,char *c)
{
	if (!dialog) return 0;
	if (!strcmp(dialog,"-")) {
		int z=g_utf8_get_char(c);
		if (z == '-' || (z >= 0x2012 && z <= 0x2015)) return 1;
		return 0;
	}
	return !strncmp(c,dialog,strlen(dialog));
}

void add_para(int pau,char *c)
{
	if (!pars) pars=malloc(sizeof(*pars) * (para_size=1024));
	else if (para_count >= para_size) {
		para_size = 2* para_size;
		pars=realloc(pars,sizeof(*pars) * para_size);
	}
    pars[para_count].par_pause=pau;
    pars[para_count++].par_body=c;
}

int last_partyp;
int in_sentence;

void push_para(char *c,int sentence)
{
	int ptyp,pau;
	if (in_sentence) {
		add_para(PAUSE_SENTENCE,c);
		in_sentence=sentence;
		return;
	}
	in_sentence=sentence;
	if (audiobook) {
		ptyp=PAR_NORMAL;
		if (titleline && para_count) ptyp=PAR_NORMAL;
		else if (is_dialog(dialog,c)) ptyp=PAR_DIALOG;
		
		if (last_partyp == PAR_EMPTY) pau=PAUSE_LONG;
		else if (!last_partyp) pau=0;
		else if (ptyp == PAR_DIALOG && last_partyp == PAR_NORMAL) pau=PAUSE_PRE;
		else if (ptyp == PAR_NORMAL && last_partyp == PAR_DIALOG) pau=PAUSE_POST;
		else if (titleline && para_count == 1) pau=PAUSE_VLONG;
		else pau=PAUSE_NORMAL;
		last_partyp=ptyp;
		
	}
	else pau=PAUSE_NORMAL;
	add_para(pau,c);
}

void add_line(char *c)
{
	char *d;
	c=trim(c);
	if (!*c) {
		if (last_partyp) last_partyp=PAR_EMPTY;
		in_sentence=0;
		return;
	}
	if (!autodivide) {
		d=c+strlen(c)-1;
		if (*d == '\\' ) {
			*d=0;
			push_para(c,1);
		}
		else {
			push_para(c,0);
		}
		return;
	}
	for (;c && *c;c=d) {
		d=c;
		for (;;) {
			d=strpbrk(d,".?!");
			if (!d) break;
			d++;
			if (g_unichar_isspace(g_utf8_get_char(d))) break;
		}
		if (d) {
			*d++=0;
			d=trim(d);
			if (!*d) {
				push_para(c,0);
				break;
			}
			push_para(c,1);
			continue;
		}
		push_para(c,0);
		break;
	}
}



int main(int argc, char *argv[])
{
    static char *optstr="hv:s:p:d:ate:rb:oxyflg:w";
    int opt,idx,ewas;
    char *c,*d,*err=NULL;
    short *last_wave=NULL,*this_wave=NULL;
    int offset;
    int count,i;
    int rest,ssml=0,filessml=0;
    int last_rest_offset,last_rest;
    int wave_open=0,sync_mode=0;
    char *debug=NULL;
    
    voices=SAPI_voices();
    if (!voices || !*voices) {
        fprintf(stderr,"No SAPI voices declared\n");
        exit(1);
    }
    double darg(double minval, double maxval) {
        char *d;
        double val=strtod(optarg,&d);
        if (d == optarg || val < minval || val > maxval) helpme(argv[0],1);
        return val;
    }
    while ((opt = getopt(argc,argv,optstr)) != -1) {
        switch(opt) {
            case 'h': helpme(argv[0],0);break;
            case 'l': ssml=1;break;
            case 'f': filessml=1;break;
            case 'v': voice=strdup(optarg);break;
            case 's': speed=darg(0.5,2.0);break;
            case 'p': pause_factor=darg(0.2,1.0);break;
            case 'b': breath=darg(0.25,2.0);break;
            case 'a': audiobook=1;break;
            case 'd': dialog=strdup(optarg);audiobook=1;break;
            case 't': titleline=audiobook=1;break;
            case 'r': is_raw=1;break;
            case 'o': do_play=1;break;
            case 'e': endline=strdup(optarg);audiobook=1;break;
            case 'x': if (autodivide) helpme(argv[0],1); sentenced=1;break;
            case 'y': if (sentenced) helpme(argv[0],1); autodivide=1;break;
            case 'g': debug=strdup(optarg);break;
	    case 'w': sync_mode=1;break;
            default: helpme(argv[0],1);
        }
    }
    if (!voice) voice=voices[0];
    else {
        for (i=0;voices[i];i++) if (!strcasecmp(voices[i],voice)) break;
        if (!voices[i]) helpme(argv[0],0);
    }
    if (debug) {
	fdebug=fopen(debug,"w");
	if (!fdebug) {
	    perror(debug);
	    exit(1);
	}
    }
    idx=optind;
    if (idx >= argc || !strcmp(argv[idx],"-")) {
        int bsize=1024;
        int rsize=0;
        if (idx < argc) idx++;
	_dbg("reading from stdin\n");
        body=malloc(bsize);
        for (;;) {
            int d=read(0,body+rsize,bsize-rsize);
            if (d <= 0) break;
            rsize+=d;
            if (rsize == bsize) {
                bsize = 2*bsize;
                body=realloc(body,bsize);
            }
        }
        body[rsize]=0;
	_dbg("read %d bytes\n",rsize);
    }
    else {
        char *fname=argv[idx++];
        struct stat sb;
        int fd;
        if (stat(fname,&sb)) {
            perror(fname);
            exit(1);
        }
        fd=open(fname,O_RDONLY);
        if (fd < 0) {
            perror(fname);
            exit(1);
        }
        body=malloc(sb.st_size+1);
        if (read(fd,body,sb.st_size) != sb.st_size) {
            perror(fname);
            exit(1);
        }
        close(fd);
        body[sb.st_size]=0;
	_dbg("Read %d bytes from %s\n",(int)(sb.st_size),fname);
    }
    if (!filessml) {
	para_count=0;
        ewas=0;
        for (c=body;c && *c;c=d) {
            int pt,pau;
            d=strpbrk(c,"\r\n");
            if (d) {
		int z=*d;
		*d++=0;
		if (z == '\r' && *d=='\n') d++;
	    }
            c=trim(c);
            add_line(c);
        }
        if (!para_count) {
            fprintf(stderr,"Nothing to say\n");
            exit(1);
        }
        if (endline) {
    		add_para(PAUSE_VLONG,endline);
        }
    }
    else {
	body=trim(body);
	if (!*body) {
            fprintf(stderr,"Nothing to say\n");
            exit(1);
	}
    }
    vox=SAPI_open(voice,&err);
    if (!vox) {
        fprintf(stderr,"Error: %s\n",err);
        exit(1);
    }
    if (fdebug) SAPI_debugFile(vox,fdebug);
    SAPI_enableSSML(vox,ssml);
    if (filessml) {
	pars=malloc(sizeof(*pars));
	pars[0].par_pause=0;
	pars[0].par_body=body;
	para_count=1;
    }
    _dbg("Found %d paragraphs\n",para_count);
    for (i=0;i<para_count;i++) {
        int pau;
        if (!last_wave || !pars[i].par_pause) pau=0;
        else {
			pau=(vox->freq * _pauses[pars[i].par_pause] * breath)/1000;
			if (pau < vox->freq/5) pau=vox->freq/5;
		}
        this_wave=SAPI_getSentence(vox,pars[i].par_body,&offset,&count,&rest,pause_factor,&err);
        if (!this_wave) {
	    _dbg("Error from SAPI subsystem, finishing\n");
            fprintf(stderr,"Error: %s\n",err);
            if (last_wave && wave_open) {
				push_wave(last_wave,last_rest_offset,last_rest);
			}
			finalize_wave();
            exit(1);
        }
	_dbg("Got wave %d - %d - %d\n",offset,count,rest);
	if (!wave_open) {
	    open_wave(vox->freq,(idx < argc)?argv[idx]:NULL);
	    wave_open=1;
	}
        if (pau) {
            push_pause(last_wave+last_rest_offset,last_rest,this_wave,offset,pau);
        }
        if (!i) {
            count += offset;
            offset=0;
        }
        push_wave(this_wave,offset,count);
	if (sync_mode) sync_wave();
        if (last_wave) free(last_wave);
        last_wave=this_wave;
        last_rest_offset = offset+count;
        last_rest = rest;
    }
    push_wave(this_wave,offset+count,rest);
    finalize_wave();
    exit(0);
}

